/**********************************************************************/
/*                                                                    */
/* DESCRIPTION: A small utility to join (interleave) files.           */
/*                                                                    */
/* INSTALLATION: gcc -O -o jj jj.c ; strip jj                         */
/*                                                                    */
/* BUGS / FEATURES:                                                   */
/* - will not accept filenames that begin with "-" or "%"             */
/* - only checks the first letter of the long-format options          */
/* - requires a space after the delim and output options              */
/* - will accept standard input multiple times                        */
/* - will accept the output file as an input file                     */
/* - does not close files before exiting                              */
/* - cannot loop standard input                                       */
/* - many text files end with "\n", the default delimiter             */
/*                                                                    */
/**********************************************************************/

#define SYNTAX "\
USAGE: %s -tlfah [-1 DELIM] [-2 DELIM] [-o FILE] [INPUT...]\n
  -t, --truncate        stop at the end of the shortest file
  -l, --loop            loop short file(s)
  -f, --flush           do not buffer output
  -a, --all       	print all delimiters
  -h, --help            print this message
  -1, --1 DELIM         input delimiter(s) (default \"\\n\")
  -2, --2 DELIM         output delimiter(s) (default \"\\n\")
  -o, --output FILE     output to FILE\n
Each INPUT is the name of a file, or a string prefixed by \"%%\".
A string INPUT is equivalent to a file consisting of that string.
The N-th INPUT is delimited by the N-th character of DELIM.
If there are more INPUTs than characters in DELIM the last
character of DELIM is repeated as neccessary.
If output delimiters are not given the input delimiters are used.
The INPUT \"-\" can be used to read from the standard input.
The \"truncate\" option overrides the \"loop\" option.\n"

#include <stdio.h>

typedef struct
{ FILE *file; char *string;
  int pos, eof, in_delim, out_delim;
  int (*fn)(void *);
} INPUT;

void *jj_malloc(size_t size);
FILE *jj_fopen(char *name, char *mode);
INPUT *munch_input(char *name);
int input_file(void *in_ptr);
int input_string(void *in_ptr);

char truncate, loop, flush, all;

int main(int argc, char **argv)
{ int c, i, j, k;
  int num_inputs; INPUT **input, *in; FILE *output;
  char *in_delim = "\n", *out_delim = 0;
  num_inputs = 0; output = stdout;
  input = (INPUT **) jj_malloc((--argc) * sizeof(INPUT *));
  truncate = loop = flush = all = 0;

  for(i = 0; i < argc; i++)
  { if(**(++argv) == '-' && (c = *(++(*argv))))
    { if(c == '-') *((++(*argv))+1) = 0;
      while(**argv) switch(*((*argv)++))
      { case 't': truncate = 1; break;
        case 'l': loop = 1; break;
        case 'a': all = 1; break;
        case 'f': flush = 1; break;
        case '1': if(!(**argv) && ++i < argc)
                  { in_delim = *(++argv); goto next; }
        case '2': if(!(**argv) && ++i < argc)
                  { out_delim = *(++argv); goto next; }
        case 'o': if(!(**argv) && ++i < argc)
                  { output = jj_fopen(*(++argv), "w"); goto next; }
        default: printf(SYNTAX, *(argv - i - 1)); return(0);
      }
    }
    else *(input + (num_inputs++)) = munch_input(*argv);
    next:
  }
  if(!num_inputs) { printf(SYNTAX, *(argv - argc)); return(0); }
  if(!out_delim) out_delim = in_delim;
  for(i = j = k = 0; k < num_inputs; k++)
  { (*(input + k))->in_delim = *(in_delim + i);
    (*(input + k))->out_delim = *(out_delim + i);
    if(! *(in_delim + (++i))) i--;
    if(! *(out_delim + (++j))) j--;
  }
  for(i = 0, j = num_inputs; j; i++, i %= num_inputs)
  { in = *(input + i);
    if((k = !in->eof) || loop)
    { while((c = in->fn(in)) != in->in_delim)
        fputc(c, output);
      if(in->eof && (j -= k) && truncate) j = 0;
    }
    if(i == (num_inputs - 1) || all)
    { fputc(in->out_delim, output);
      if(flush) fflush(output);
    }
  }
  return(1);
}

INPUT *munch_input(char *name)
{ INPUT *in = (INPUT *) jj_malloc(sizeof(INPUT));
  in->pos = in->eof = 0;
  if (*name == '%')
  { in->fn = &input_string;
    in->string = name + 1;
  }
  else
  { in->fn = &input_file;
    in->file = jj_fopen(name, "r");
  }
  return(in);
}

int input_file(void *in_ptr)
{ int c; INPUT *in = (INPUT *) in_ptr;
  c = fgetc(in->file);
  if(c == EOF)
  { if(loop && in->file != stdin) rewind(in->file);
    in->eof = 1;
    c = in->in_delim;
  }
  return(c);
}

int input_string(void *in_ptr)
{ int c; INPUT *in = (INPUT *) in_ptr;
  c = *(in->string + (in->pos++));
  if(!c)
  { if(loop) in->pos = 0;
    in->eof = 1;
    c = in->in_delim;    
  }
  return(c);
}

void *jj_malloc(size_t size)
{ void *ptr;
  if(! (ptr = (void *)malloc(size)))
  { fprintf(stderr, "Memory allocation error.\n");
    exit(0);
  }
  return(ptr);
}

FILE *jj_fopen(char *name, char *mode)
{ FILE *file;
  if(! (*name)) return(stdin);
  if(! (file = fopen(name, mode)))
  { fprintf(stderr, "Could not open file %s.\n", name);
    exit(0);
  }
  return(file);
}
